#!/usr/bin/perl

# GPV (Graphical Pipeline Viewer) is unsupported software.
# (c) Feb 2000 
# Kenneth C. Barr 
# Eric D. Marsman
#
# A Pipeline Visualization tool for use with Simplescalar 3.0.  
#
# Written in Perl/Tk, this tool is intended to be a portable, graphical replacement for simplesim-3.0/pipeview.pl.  It was 
# written at the University of Michigan for Todd Austin's Microarchitecture course by Ken Barr and Eric Marsman
#
# You must download Perl/Tk, too.      


use Tk;
use Getopt::Std;

#constants
$COLORFILE="colors.cfg";
$NUMCOLORS=10;            #number of colors to expect in colors.cfg file
$TCLK=1;                  #width of a clock cycle in pixels
$YSPACE=2;                #number of pixels between each line
$SCALEHEIGHT=20;          #height in pixels of the scale.  (so we don't overlap graph on the scale)
$THICKNESS=1;             #how thick is a line?

$CANVASWIDTH=800;         #default canvas width
$CANVASHEIGHT=600;        #height
$INITIALDIR=`pwd`;        #or $ENV{HOME}, whatever
$BLINK=10000;             #number of loop iterations between blinking rectangles


#types of files to display in popup windows
$tracetypes=[
            ['Trace Files', '.trc',],
            ['All Files','*',],
            ];
$pstypes=[
            ['Postscript Files', '.ps',],
            ['All Files','*',],
            ];

$XPAD= 50;            #default padding for dragging
$YPAD= 50;

$HELVBOLD="-adobe-helvetica-bold-r-normal--12-120-75-75-p-70-hp-roman8";

#globals
my $usage=0;		  #print the usage message?
my($fname);               #filename for file dialog
$i=0;
%color;                   #hash of colors
$parsing=1;               #used so we don't modify screen until its drawn
$scale=1;                
$zoomcount=0;             #increment on zin, decrement on zout
$item_cnt=0;              #total number of items on screen... do not reset on add trace :)


getopts('w:h:c:t:y:s:x:b:u', \%opts);  # each switch is optional.  Values in %opts
#if options are specified, then overwrite the defaults.

if($opts{w}){ $CANVASWIDTH=$opts{w}; }
if($opts{h}){ $CANVASHEIGHT=$opts{h}; }
if($opts{c}){ $COLORFILE=$opts{c}; }
if($opts{t}){ $TRC_FILE=$opts{t}; }
if($opts{y}){ $THICKNESS=$opts{y}; }
if($opts{s}){ $YSPACE=$opts{s}; }
if($opts{x}){ $TCLK=$opts{x}; }
if($opts{b}){ $BLINK=$opts{b}; }
if($opts{u}){ $usage=1; }

if(@ARGV || $usage==1){
      print STDERR "Usage: gpv.pl\t[-c <color file>] [-w canvaswidth] [-h canvasheight]\n";
	  print STDERR "\t\t[-t <trace file>] [-y line thickness] [-s vertical line spacing]\n";
	  print STDERR "\t\t[-x cycle width] [-b blink delay]\ngpv -u prints this message\n";
      exit -1;
}

&parseColorFile;
&drawWindow;

#if a file was specified with -t, parse it.
if($TRC_FILE) {
      ptrace($TRC_FILE);
}

#wait for user input and interaction
MainLoop;

sub ptrace{
      my($tracefile)=shift(@_);
      
      &revertscale; #just in case window has changed
      
      #calculate the largest windowsize to show
      if ($xdatamax > $large_xdatamax) {
            $large_xdatamax = $xdatamax;
      }
      if ($ydatamax > $large_ydatamax) {
            $large_ydatamax = $ydatamax;
      }
      
	  #erase our hash tables from previous process
	  undef %insts;
	  undef %insts_endpoint;
	  undef %insts_lastcolor;
	  undef %insts_legend;
	  undef %insts_color;

	  open(TRC_FILE, $tracefile)
		  || die "Cannot open pipeline trace file `$tracefile'";
	  
	  $cycle = 0;
	  
	  print "Instruction event legend:\n";
	  print "\n";
	  print "    * - cache miss\n";
	  print "    ! - TLB miss\n";
	  print "    / - branch misprediction\n";
	  print "    \\ - branch misprediction detected\n";
	  print "    + - address generation execution\n";
	  print "\n";
	  
	  while (<TRC_FILE>)
	  {
		  # remove carriage return
		  chop;
		  
		  # new instruction
		  if (/^\+\s+(\d+)\s+(0x[0-9a-fA-F]+)\s+(0x[0-9a-fA-F]+)\s+(.*)\n?$/)
		  {
			  # register this instruction in the live instruction hash table
			  $insts{$1} = 1;
			  $insts_pc{$1} = $2;
			  $insts_addr{$1} = $3;
			  $insts_asm{$1} = $4;
			  $insts_iid{$1} = 
                  chr ((ord("a") + (($1 / 26) % 26))) .
					  chr (ord("a") + ($1 % 26));
			  
			  $insts_endpoint{$1}=$cycle;
			  $insts_lastcolor{$1}=$if_color;
			  $insts_legend{$1}=$4;           #store asm code            
			  
			  # print instruction id and asm info
			  print "$insts_iid{$1} = `$2: $4'\n";
		  }
		  # deleted instruction
		  elsif (/^\-\s+(\d+)\n?$/)
		  {
			  # record deletion, these are processed after pipe info is printed
			  @dead_pseqs = ($1, @dead_pseqs);
		  }
		  # new cycle
		  elsif (/^@\s+(\d+)\n?$/)
		  {
            # clear all instruction stage lists
            @if_iids = ();
            @da_iids = ();
            @ex_iids = ();
            @wb_iids = ();
            @ct_iids = ();
            foreach $pseq (keys %insts){
                  if ($insts{$pseq}) {
                        $event = hex($insts_events{$pseq});
                        $longerevent = hex($insts_longerlat{$pseq});
                        #print "LONG=$longerevent\n";
                        $stage = $insts_stage{$pseq};
                        $linecolor_stage=$color{$stage};
                        if($event > 0 ){
                              #if there are multiple events, color based on the longer one.
                              if(multipleEvents($event)){
                                    $linecolor=$color{$longerevent};
                              }
                              else{ 
					     #else assign color for events
                                   if($event & 0x00000010){
                                        $linecolor=$color{16};
                                   }
                                   if($event & 0x00000008){
                                        $linecolor=$color{8};
                                   }
                                   if($event & 0x00000004){
                                         $linecolor=$color{4};
                                   }
                                   if($event & 0x00000002){
                                         $linecolor=$color{2};
                                   }
                                   if($event & 0x00000001){
                                         $linecolor=$color{1};
                                   }
                              }
                        }
                        else{
                              $linecolor = $linecolor_stage;
                        }
                        
                        if(($stage eq "IF") && ($insts_latent{$pseq}>0)){                           
                              #we need to draw a line backwards 
                              $canvas->createRectangle(
						($cycle-$insts_latent{$pseq})*$TCLK, 
						$pseq*($YSPACE+$THICKNESS), 
						$cycle*$TCLK+$TCLK, 
						$pseq*($YSPACE+$THICKNESS)+$THICKNESS, 
						-fill=>$linecolor, 
						-outline=>$linecolor, 
						-tag=>$linecolor);          
                              $item_cnt++;
                              $insts_color{$pseq} = $linecolor;
                              $insts_endpoint{$pseq}=$cycle*$TCLK+$TCLK;
                        }
                        elsif($insts_color{$pseq} eq $linecolor) {
                               #same color, don't draw yet, keep old endpoint and color
                        }
                        else {
                              #draw a line
                              $canvas->createRectangle(
						$insts_endpoint{$pseq}, 
						$pseq*($YSPACE+$THICKNESS), 
						$cycle*$TCLK, 
						$pseq*($YSPACE+$THICKNESS)+$THICKNESS,
						-fill=>$insts_color{$pseq}, 
						-outline=>$insts_color{$pseq}, 
						-tag=>$linecolor);
                                    
                              $canvas->createRectangle(
						$cycle*$TCLK, 
						$pseq*($YSPACE+$THICKNESS), 
						$cycle*$TCLK+$TCLK, 
						$pseq*($YSPACE+$THICKNESS)+$THICKNESS, 
						-fill=>$linecolor, 
						-outline=>$linecolor);           
                              $item_cnt++;
                              $item_cnt++;
                              $insts_color{$pseq} = $linecolor;
                              $insts_endpoint{$pseq}=$cycle*$TCLK+$TCLK;
                        }
                        $ydatamax=$pseq*($YSPACE+$THICKNESS)+$THICKNESS;
                        $xdatamax=$cycle*$TCLK+$TCLK;
             
                        #comment this out for debugging to speed things along
                        $canvas->update;
                  }
                  else {
                        print "a dead instruction here!  That's unexpected\n";
                        exit -1;
                  }
            }
            # print last cycle pipeline state
            foreach $pseq (keys %insts) {
                  # partition by instruction stage
                  $stage = $insts_stage{$pseq};

                  # decode the instruction events
                  $events = hex($insts_events{$pseq});
                  $evstr = "";
                  if ($events & 0x00000001) {              # cache miss
                        $evstr = $evstr . "*";
                  }
                  if (($events & 0x00000002) != 0) {       # TLB miss
                        $evstr = $evstr . "!";
                  }
                  if ($events & 0x00000004) {              # mis-predict occurred
                        $evstr = $evstr . "/";
                  }
                  if ($events & 0x00000008) {              # mis-predict detected
                        $evstr = $evstr . "\\";
                  }
                  if ($events & 0x00000010) {              # addr generation
                        $evstr = $evstr . "+";
                  }

                  if ($stage eq "IF") {
                        @if_iids = (@if_iids, $insts_iid{$pseq} . $evstr);
                  }
                  elsif ($stage eq "DA") {
                        @da_iids = (@da_iids, $insts_iid{$pseq} . $evstr);
                  }
                  elsif ($stage eq "EX") {
                        @ex_iids = (@ex_iids, $insts_iid{$pseq} . $evstr);
                  }
                  elsif ($stage eq "WB") {
                        @wb_iids = (@wb_iids, $insts_iid{$pseq} . $evstr);
                  }
                  elsif ($stage eq "CT") {
                        @ct_iids = (@ct_iids, $insts_iid{$pseq} . $evstr);
                  }
                  else {
                        print "warning: unknown stage\n";
                  }
            }

#       print "** if_iids: ";
#       foreach $iid (@if_iids) {
#           print "$iid ";
#       }
#       print "\n";

            # sort stage lists
            @if_iids = sort @if_iids;
            @da_iids = sort @da_iids;
            @ex_iids = sort @ex_iids;
            @wb_iids = sort @wb_iids;
            @ct_iids = sort @ct_iids;

            # print pipeline header
            if ((@if_iids + @da_iids + @ex_iids + @wb_iids + @ct_iids) > 0) {
                  print "\n";
                  print " [IF]     ", " [DA]     ", " [EX]     ", " [WB]     ", " [CT]", "\n";
            }

            # print stage data
            while ((@if_iids + @da_iids + @ex_iids + @wb_iids + @ct_iids) > 0) {
                  if (@if_iids > 0) {
                        ($iid, @if_iids) = @if_iids;
                        printf "  %-6s  ", $iid;
                  }
                  else {
                        print "          ";
                        
                  }

                  if (@da_iids > 0) {
                        ($iid, @da_iids) = @da_iids;
                        printf "  %-6s  ", $iid;
                  }
                  else {
                        print "          ";
                  }

                  if (@ex_iids > 0) {
                        ($iid, @ex_iids) = @ex_iids;
                        printf "  %-6s  ", $iid;
                  }
                  else {
                        print "          ";
                  }

                  if (@wb_iids > 0) {
                        ($iid, @wb_iids) = @wb_iids;
                        printf "  %-6s  ", $iid;
                  }
                  else {
                        print "          ";
                  }

                  if (@ct_iids > 0) {
                        ($iid, @ct_iids) = @ct_iids;
                        printf "  %-6s  ", $iid;
                  }
                  else {
                        print "          ";
                  }
                  print "\n";
            }
            print "\n";

            # delete instruction that finished in this cycle
            foreach $pseq (@dead_pseqs) {
                  delete $insts{$pseq};
                  delete $insts_pc{$pseq};
                  delete $insts_addr{$pseq};
                  delete $insts_asm{$pseq};
                  delete $insts_iid{$pseq};
                  delete $insts_stage{$pseq};
                  delete $insts_events{$pseq};
                  delete $insts_latent{$pseq};
                  delete $insts_longerlat{$pseq}
            }
            @dead_pseqs = ();

      # go on to the next cycle
            $cycle = $1;
            print "@ $cycle\n";
      }
    # instruction transition to a new stage
    elsif (/^\*\s+(\d+)\s+(\w+)\s+(0x[0-9a-fA-F]+)\s+(\d+)\s+(0x[0-9a-fA-F]+)\n?$/) {
            if ($insts{$1}) {
                  # this instruction is live...
                  $insts_stage{$1} = $2;
                  $insts_events{$1} = $3;
                  $insts_latent{$1} = $4;
                  $insts_longerlat{$1} = $5;
            }
            else {
                  # this instruction is dead...
                  # print STDERR "inst `$1' is dead...\n";
            }
      }
    else {
            print "warning: could not parse line: `$_'\n";
      }
}
#done drawing
#check to see if new trace is larger than old trace
if($xdatamax < $large_xdatamax) {
      $xdatamax = $large_xdatamax;
}
if($ydatamax < $large_ydatamax) {
      $ydatamax = $large_ydatamax;
}

#now confine the canvas
$canvas->configure(
	-scrollregion => [-$XPAD,
				-$YPAD,(($xdatamax+$XPAD)*$scale), 
				(($ydatamax+$YPAD)*$scale)], 
				confine=>1);
$scale_canvas->configure(
	-scrollregion=>  [-$XPAD,
				0,(($xdatamax+$XPAD)*$scale), 
				$SCALEHEIGHT], 
				confine=>1);
$canvas->update;
&drawScale(1, $xdatamax);

$parsing=0;

close(TRC_FILE);
}

sub parseColorFile{
my @color;

      #parse the color configuration file
      open(CLRZ, $COLORFILE) || die "Cannot open color configuration file `$COLORFILE'";
      
      while (<CLRZ>) {
            if (/^(\w+)\s\n?$/) {
                  #found color
                  $color[$i]= $1;
                  $i++;
            }
            elsif (/^\#.*\s\n?$/) {
                  #ignore comments
            }
            else {
                  print STDERR "error in `$COLORFILE'. Unmatched string.\n";
                  exit -1;
            }
      }
      close(CLRZ);
      if($i!=$NUMCOLORS){
            print "Not enough colors in $COLORFILE.  Expecting $NUMCOLORS\n";
            exit -1;
      }
      
      $color{"IF"} = $color[0];
      $color{"DA"} = $color[1];
      $color{"EX"} = $color[2];
      $color{"WB"} = $color[3];
      $color{"CT"} = $color[4];
      $color{1}    = $color[5];  #cache miss
      $color{2}    = $color[6];  #TLB miss
      $color{4}    = $color[7];  #branch mispredict
      $color{8}    = $color[8];  #mis-predict detect
      $color{16}   = $color[9];  #address generation
}

sub drawWindow{

      $mw= new MainWindow;
      $mw->title("SimpleScalar - Graphical Pipeline Viewer");

$DIALOG_ABOUT=$mw->Dialog('-bitmap' => 'info', 
                              -default_button => 'OK', 
                              -buttons=>['OK'], 
                              -title => 'Simplescalar GPV',
                              -text => "(c) Feb 2000\nKenneth C. Barr\nEric D. Marsman\n\nPipeline Visualization tool for use with Simplescalar 3.0.  Written in Perl/Tk, this tool is intended to be a portable, graphical replacement for simplesim-3.0/pipeview.pl.  It was written at the University of Michigan for Todd Austin's Microarchitecture course by Ken Barr and Eric Marsman\n\nGo Blue!");

      $menuframe=$mw->Frame(
		-relief=>"ridge", 
		-borderwidth=>2)
		->pack(
			-side=>'top', 
			-anchor=>'n', 
			-fill=>'x');
      $menuframe->Menubutton(
		-text=>"File", 
		-tearoff=>0, 
		-menuitems=> [['command' => "Add trace...", 
                -command=> sub {
      		$fname = $mw->getOpenFile(
					-filetypes=>$tracetypes, 
					-initialdir=>$INITIALDIR, 
					-title=>"Open trace file");
              	#print("FILEDIALOG returned $fname\n");
              	if (defined($fname)) {
                    	&ptrace($fname);
              	}
     		}],
    		['command' => "Print...",
     		-command=>sub {
              	$fname = $mw->getSaveFile(
					-filetypes=>$pstypes, 
					-initialdir=>$ENV{HOME}, 
					-title=>"Print to postscript file");
          		if (defined($fname)) {
                	&print($fname);
          		}
   		}],
 		"-",
  		['command' => "Exit", -command=>sub{exit}]]
       	)->pack(-side=>'left');

      $menuframe->Menubutton(
		-text=>"View", 
		-tearoff=>0, 
		-menuitems=> [['command' => "Zoom In", 
					-command => \&zin],
           			['command' => "Zoom Out", 
					-command => \&zout],
           			"-",
           			['command' => "Revert Scale", 
					-command => \&revertscale]
           	]
      	)->pack(-side=>'left');

      $menuframe->Menubutton(
		-text=>"Help", 
		-tearoff=>0, 
		-menuitems=> [
             ['command' => "About", -command=>[$DIALOG_ABOUT => 'Show']],
		]
      	)->pack(-side=>'right');

      #exit; don't do cleanup.   Use destroy if cleanup necessary.
      
      #pack buttons along the top
      $toolbarframe=$mw->Frame(
		-relief=>"ridge", 
		-borderwidth=>2)
		->pack(
			-side=>'top',
			 -expand=>1, 
			 -fill=>'x', 
			 -anchor=>'n');
      $zin_b=$toolbarframe->Button(
		-text => "Zoom In",       
		-command => \&zin)
		->pack(
			-side=>'left', 
			-ipadx=>10, 
			-anchor=>'n', 
			-expand=>1);       
      $zout_b=$toolbarframe->Button(
		-text => "Zoom Out",      
		-command => \&zout)
		->pack(
			-side=>'left', 
			-ipadx=>10,
			-anchor=>'n', 
			-expand=>0);
    	$revert_b=$toolbarframe->Button(
		-text => "Revert Scale",  
		-command => \&revertscale)
		->pack(
			-side=>'left',
			-ipadx=>10, 
			-padx=>200,
			-anchor=>'n',
			-expand=>1);
     
#$process_b=$toolbarframe->Button(-text => "Process Trace", -command => \&ptrace)->pack(-side=>'left',-ipadx=>10,-anchor=>'n', -expand=>1);
#$exit_b=$toolbarframe->Button(-text => "Exit", -command => sub{exit})->pack(-side=>'left',-ipadx=>10,-anchor=>'n',-expand=>1);

      $legend=$mw->Canvas(
		-background=>'white', 
		-width=>$CANVASWIDTH/4)
		->pack(
			-side=>'right', 
			-anchor=>'n', 
			-fill=>'y');
      $sby=$mw->Scrollbar(
		-orient=>'vertical')
		->pack( 
			-side=>'right', 
			-anchor=>'w', 
			-expand=>1, 
			-fill=>'y');
      $sbx=$mw->Scrollbar(-orient=>'horizontal');
      $scale_canvas=$mw->Canvas(
		-height=>$SCALEHEIGHT, 
            -background=>'white',
            -xscrollincrement=>10,
            -yscrollincrement=>10,
            -xscrollcommand=>['set' => $sbx]
            )->pack(
			-anchor=>'s', 
			-expand=>1, 
			-fill=>'x');

      $sbx->pack( 
		-side=>'bottom', 
		-expand=>1, 
		-anchor=>'n', 
		-fill=>'x');

      $canvas= $mw->Canvas(
            -background=> 'white',
            -height=>$CANVASHEIGHT,
            -width=>$CANVASWIDTH,
            -xscrollincrement=>10,
            -yscrollincrement=>10,
            -xscrollcommand=>['set' => $sbx],
            -yscrollcommand=>['set' => $sby]
            )->pack(
			-side=>'left');

      #setup the scroll bars
      $sbx->configure(-command=>\&myXscroll);
      $sby->configure(-command=>['yview' => $canvas]);

      #draw the legend
      $legend->createRectangle(10,10,20,20, -fill=>$color{IF});
      $legend->createRectangle(10,30,20,40, -fill=>$color{DA});
      $legend->createRectangle(10,50,20,60, -fill=>$color{EX});
      $legend->createRectangle(10,70,20,80, -fill=>$color{WB});
      $legend->createRectangle(10,90,20,100, -fill=>$color{CT});

      $legend->createRectangle(10,120,20,130, -fill=>$color{1});
      $legend->createRectangle(10,140,20,150, -fill=>$color{2});
      $legend->createRectangle(10,160,20,170, -fill=>$color{4});
      $legend->createRectangle(10,180,20,190, -fill=>$color{8});
      $legend->createRectangle(10,200,20,210, -fill=>$color{16});
      
      $legend->createText(20, 15, -text=>"  IF", -anchor=>'w');
      $legend->createText(20, 35, -text=>"  DA", -anchor=>'w');
      $legend->createText(20, 55, -text=>"  EX", -anchor=>'w');
      $legend->createText(20, 75, -text=>"  WB", -anchor=>'w');
      $legend->createText(20, 95, -text=>"  CT", -anchor=>'w');
      $legend->createText(20, 125, -text=>"  Cache Miss", -anchor=>'w');
      $legend->createText(20, 145, -text=>"  TLB Miss", -anchor=>'w');
      $legend->createText(20, 165, -text=>"  Mispred occured", -anchor=>'w');
      $legend->createText(20, 185, -text=>"  Mispred detected", -anchor=>'w');
      $legend->createText(20, 205, -text=>"  Address Generation", -anchor=>'w');

      $legend->createText(
		10, 
		250, 
		-font=>$HELVBOLD, 
		-text=>"Button1 - Highlight Event", 
		-anchor=>'w');
      $legend->createText(
		10, 
		270, 
		-font=>$HELVBOLD, 
		-text=>"Button2 - Drag Canvas", 
		-anchor=>'w');
      $legend->createText(
		10, 
		290, 
		-font=>$HELVBOLD, 
		-text=>"Button3 - Show Assembly", 
		-anchor=>'w');


#routines for the canvas
 $legend->Tk::bind("<1>", [ \&flash, Ev('x'), Ev('y') ]);
 sub flash {
     if($parsing) {
           return;
     }
     my ($canv, $x, $y) = @_;
     $x=$canv->canvasx($x);
     $y=$canv->canvasy($y);
     if(($y >= 10) && ($y < 20)) {
                 #print "IF\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{IF}) {
                             $canvas->itemconfigure($i, -fill=>black, -outline=>black);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{IF}, -outline=>$color{IF});
                             $canvas->update;
                       }
                 }
           }
      }
      if(($y > 30) && ($y < 40)) {
                 #print "DA\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{DA}) {
                             $canvas->itemconfigure($i, -fill=>black, -outline=>black);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{DA}, -outline=>$color{DA});
                             $canvas->update;
                       }
                 }
           }
      }
      if(($y > 50) && ($y < 60)) {
                 #print "EX\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{EX}) {
                             $canvas->itemconfigure($i, -fill=>black, -outline=>black);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{EX}, -outline=>$color{EX});
                             $canvas->update;
                       }
                 }
           }
      }
      if(($y > 70) && ($y < 80)) {
                 #print "WB\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{WB}) {
                             $canvas->itemconfigure($i, -fill=>magenta, -outline=>magenta);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{WB}, -outline=>$color{WB});
                             $canvas->update;
                       }
                 }
           }
      }
      if(($y > 90) && ($y < 100)) {
                 #print "CT\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{CT}) {
                             $canvas->itemconfigure($i, -fill=>magenta, -outline=>magenta);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{CT}, -outline=>$color{CT});
                             $canvas->update;
                       }
                 }
           }
      }
      if(($y > 120) && ($y < 130)) {
                 #print "cache\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{1}) {
                             $canvas->itemconfigure($i, -fill=>magenta, -outline=>magenta);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{1}, -outline=>$color{1});
                             $canvas->update;
                       }
                 }
           }
      }
      if(($y > 140) && ($y < 150)) {
                 #print "TLB\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{2}) {
                             $canvas->itemconfigure($i, -fill=>magenta, -outline=>magenta);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{2}, -outline=>$color{2});
                             $canvas->update;
                       }
                 }
           }
      }
      if(($y > 160) && ($y < 170)) {
                 #print "mpred\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{4}) {
                             $canvas->itemconfigure($i, -fill=>magenta, -outline=>magenta);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{4}, -outline=>$color{4});
                             $canvas->update;
                       }
                 }
           }
      }
      if(($y > 180) && ($y < 190)) {
                 #print "mpredet\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{8}) {
                             $canvas->itemconfigure($i, -fill=>magenta, -outline=>magenta);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{8}, -outline=>$color{8});
                             $canvas->update;
                       }
                 }
           }
      }
      if(($y > 200) && ($y < 210)) {
                 #print "addr\n";
            for($k=0; $k < 1; $k++) {
                 for ($i = 0; $i < $item_cnt; $i++) {
                       $current_color=$canvas->itemcget($i, -fill);
                       if($current_color eq $color{16}) {
                             $canvas->itemconfigure($i, -fill=>magenta, -outline=>magenta);
                             $canvas->update;
                             for($j=0; $j < $BLINK; $j++){}
                             $canvas->itemconfigure($i, -fill=>$color{16}, -outline=>$color{16});
                             $canvas->update;
                       }
                 }
           }
      }
      
     
 }


$canvas->Tk::bind("<3>", [ \&inst_legend, Ev('x'), Ev('y')]);
sub inst_legend {
      my ($canv, $x, $y) = @_;
      if($parsing) {
            return;
      }

      $inst_cycle= ($canv->canvasx($x))/$TCLK;
      $pseq_index= (($canv->canvasy($y))/($YSPACE + $THICKNESS));
      #$pseq_index= ($canv->canvasy($y) * ($YSPACE + $THICKNESS));
      
      $inst_cycle= int ($inst_cycle / $scale);
      $pseq_index= int ($pseq_index / $scale);
      
      #clear old inst
      $legend->delete("asm2","asm4");

      $legend->createText(10, 340, -font=>$HELVBOLD, -text=>"cycle: ", -anchor=>'w', -tag=>"asm1");
      $legend->createText(10, 360, -text=>$inst_cycle, -anchor=>'w', -tag=>"asm2");
      $legend->createText(10, 385, -font=>$HELVBOLD, -text=>"inst: ", -anchor=>'w', -tag=>"asm3");
      $legend->createText(10, 405, -text=>$insts_legend{$pseq_index}, -anchor=>'w', -tag=>"asm4");
}


$canvas->CanvasBind('<2>' => 
sub {
      my ($canvas)= @_;
      if($parsing){
            return;
      }
      my $e= $canvas->XEvent;
      $canvas->scan('mark', $e->x, $e->y);
      $scale_canvas->scan('mark', $e->x, $e->y);
      #canvas is confined initally and in zoom in/out functions
});

$canvas->CanvasBind('<B2-Motion>' => 
sub {
      my ($canvas)= @_;
      
      my $e = $canvas->XEvent;
      $canvas->scan('dragto', $e->x, $e->y);
      $scale_canvas->scan('dragto', $e->x, $e->y);
      #canvas is confined initally and in zoom in/out functions

});

&drawScale(1, $CANVASWIDTH);
} #end sub drawWindow

sub drawScale{
      my ($scale, $high)=@_;
      if($high<$CANVASWIDTH){ $high=$CANVASWIDTH; }
      $scale_canvas->delete("all");
      
      for($i=0;$i<$high;$i+=10){
            if($i % 100 == 0){
                  $scale_canvas->createText(
				$i*$scale*$TCLK, 
				5, 
				-text=>$i, 
				-anchor=>"n");
            }
            else{
                  $scale_canvas->createLine(
				$i*$scale*$TCLK, 
				0, 
				$i*$scale*$TCLK, 
				5, 
				-fill=>'black');
            }
      }       
      $scale_canvas->update;
}

sub print {
      my $psfile=shift(@_);
      if($parsing) {
            return;
      }
      $postscript= $canvas->postscript();
      $canvas->postscript(-file=> $psfile);
}

sub revertscale{
      if($parsing) {
            return;
      }
      if($zoomcount > 0){
            while($zoomcount){ &zout; }
      }
      else{
            while($zoomcount < 0){ &zin; }
      }
      myXscroll("moveto",0);
      $canvas->yview(moveto=>0);
}


sub zout {
      if($parsing) {
            return;
      }
      #$canvas->scale("all", $CANVASWIDTH/2,$CANVASHEIGHT/2, .5,.5);
      $canvas->scale("all", 0,0, .5,.5);
      $scale = $scale * .5;
 
      &drawScale($scale, $xdatamax);

      $canvas->configure(
		-scrollregion => [
			-$XPAD,
			-$YPAD,(($xdatamax+$XPAD)*$scale), 
			(($ydatamax+$YPAD)*$scale)], 
			confine=>1);
      $scale_canvas->configure(
		-scrollregion => [
			-$XPAD,
			0,
			(($xdatamax+$XPAD)*$scale),
			$SCALEHEIGHT], 
			confine=>1);
      
      $zoomcount--;
}

sub zin {
      if($parsing) {
            return;
      }
      #$canvas->scale("all", $CANVASWIDTH/2,$CANVASHEIGHT/2, 2,2);
      $canvas->scale("all", 0,0, 2,2);
      $scale = $scale * 2;
 
      &drawScale($scale, $xdatamax);
      
      $canvas->configure(
		-scrollregion => [
			-$XPAD,
			-$YPAD,
			(($xdatamax+$XPAD)*$scale), 
			(($ydatamax+$YPAD)*$scale)], 
			confine=>1);
      $scale_canvas->configure(
		-scrollregion => [
			-$XPAD,
			0,
			(($xdatamax+$XPAD)*$scale),
			$SCALEHEIGHT], 
			confine=>1);

      $zoomcount++;
}

sub myXscroll{
      #horizontally scrolls both the canvas widgets
      my ($method, @args)=@_;
         $canvas->xview($method=>@args);
         $scale_canvas->xview($method=>@args);
}

#determines whether there were multiple events (just counts the set bits)
sub multipleEvents{
      my $event=shift (@_);
      my $i=0;
      my $bits=0;
      for($i=0 ; $i<64 ; $i++){
            $bits+=($event & 0x00000001);
            $event = $event >> 1;
      }
      return ($bits > 1);
}

#$canvas->Tk::bind("<1>", [ \&scrollandzoom, Ev('x'), Ev('y') ]);
#sub scrollandzoom {
#       my ($canv, $x, $y) = @_;
#       if($parsing) {
#             return;
#       }
#
#       &zin;
#
#       #$xdist=($CANVASWIDTH/2 - $canv->canvasx($x));
#       #$ydist=($CANVASHEIGHT/2 - $canv->canvasy($y));
#
#      $xdist=$xdatamax/2;
#      $ydist=$ydatamax/2;
#
#      if($xdist < 0) {
#            $xnum = -1;
#            $xdist = -$xdist;
#      }
#      else {
#            $xnum = 1;
#      }
#      if($ydist < 0) {
#            $ynum = -1;
#            $ydist = -$ydist;
#      }
#      else {
#            $ynum = 1;
#      }
#
#      while ($xdist > 0) {
#            $canvas->xview(scroll => $xnum, "units");
#            $scale_canvas->xview(scroll => $xnum, "units");
#            $xdist = $xdist - 10;
#      }
#      while ($ydist > 0) {
#            $canvas->yview(scroll => $ynum, "units");
#            $ydist = $ydist - 10;
#      }
#
#       #myXscroll("moveto",$xfrac);
#       #$canvas->yview(moveto=>$yfrac);
#
#       #$xmovecnt = $xmovecnt + $canv->canvasx($x);
#       #$ymovecnt = $ymovecnt + $canv->canvasy($y);
#}
